"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Analyser = exports.dummyCustomSchema = void 0;
const database_engines_1 = require("../database-engines");
exports.dummyCustomSchema = {
    settings: {
        engine: database_engines_1.databaseEngines.MARIADB,
        disableTriggers: false,
        ignoredTables: [],
        tablesToFill: [],
        options: [],
        values: {}
    },
    tables: [],
};
class Analyser {
    constructor(dbConnector, customSchema, customizer, logger) {
        this.dbConnector = dbConnector;
        this.customSchema = customSchema;
        this.customizer = customizer;
        this.logger = logger;
        this.extractForeignKeys = (table) => __awaiter(this, void 0, void 0, function* () {
            const foreignKeys = yield this.dbConnector.getForeignKeys(table);
            const customTable = Object.assign({
                name: '',
                columns: [],
                maxLines: 0,
            }, this.customSchema.tables.find(t => t.name.toLowerCase() === table.name.toLowerCase()));
            for (let c = 0; c < table.columns.length; c++) {
                const column = table.columns[c];
                const match = foreignKeys.find((fk) => fk.column.toLowerCase() === column.name.toLowerCase());
                if (match) {
                    column.foreignKey = { table: match.foreignTable, column: match.foreignColumn };
                    column.options.unique = column.options.unique || match.uniqueIndex;
                }
                const customColumn = customTable.columns.find(cc => cc.name.toLowerCase() === column.name.toLowerCase());
                if (customColumn) {
                    column.options = Object.assign({}, column.options, customColumn.options);
                    if (customColumn.foreignKey)
                        column.foreignKey = customColumn.foreignKey;
                    if (customColumn.values)
                        column.values = customColumn.values;
                }
                if (column.foreignKey) {
                    table.referencedTables.push(column.foreignKey.table);
                }
            }
        });
        /** @todo: Remove deprecated warning */
        let useDeprecatedLines = false;
        customSchema.tables.forEach((table) => {
            if (table.lines) {
                useDeprecatedLines = true;
                table.maxLines = table.lines;
            }
        });
        if (useDeprecatedLines)
            logger.warn('DEPRECATED: Table.lines is deprecated, please use table.maxLines instead.');
        /****************/
    }
    analyse() {
        return __awaiter(this, void 0, void 0, function* () {
            let tables = yield this.dbConnector.getTablesInformation(this.customSchema.settings.ignoredTables, this.customSchema.settings.tablesToFill);
            tables = yield Promise.all(tables.map((table) => __awaiter(this, void 0, void 0, function* () {
                yield this.extractColumns(table);
                this.customizer.customizeTable(table);
                yield this.extractForeignKeys(table);
                return table;
            })));
            return this.generateJson(this.orderTablesByForeignKeys(tables));
        });
    }
    extractColumns(table) {
        return __awaiter(this, void 0, void 0, function* () {
            this.logger.info(table.name);
            const columns = yield this.dbConnector.getColumnsInformation(table);
            columns
                .filter((column) => {
                return ['enum', 'set'].includes(column.DATA_TYPE || '');
            }).forEach((column) => {
                column.NUMERIC_PRECISION = column.COLUMN_TYPE.match(/[enum,set]\((.*)\)$/)[1].split('\',\'').length;
            });
            table.columns = columns.map((column) => {
                const options = {
                    max: 0,
                    min: 0,
                    autoIncrement: false,
                    nullable: false,
                    unique: false,
                    unsigned: false
                };
                if (column.COLUMN_KEY && column.COLUMN_KEY.match(/PRI|UNI/ig))
                    options.unique = true;
                if (column.IS_NULLABLE === 'YES')
                    options.nullable = true;
                options.max = column.CHARACTER_MAXIMUM_LENGTH || column.NUMERIC_PRECISION;
                if (column.COLUMN_TYPE && column.COLUMN_TYPE.includes('unsigned'))
                    options.unsigned = true;
                if (column.EXTRA && column.EXTRA.includes('auto_increment'))
                    options.autoIncrement = true;
                switch (column.DATA_TYPE) {
                    case 'bit':
                        break;
                    case 'bool':
                    case 'boolean':
                        break;
                    case 'smallint':
                        if (options.unsigned) {
                            options.min = 0;
                            options.max = 65535;
                        }
                        else {
                            options.min = -32768;
                            options.max = 32767;
                        }
                        break;
                    case 'mediumint':
                        if (options.unsigned) {
                            options.min = 0;
                            options.max = 16777215;
                        }
                        else {
                            options.min = -8388608;
                            options.max = 8388607;
                        }
                        break;
                    case 'tinyint':
                        if (options.unsigned) {
                            options.min = 0;
                            options.max = 255;
                        }
                        else {
                            options.min = -128;
                            options.max = 127;
                        }
                        break;
                    case 'int':
                    case 'integer':
                    case 'bigint':
                        if (options.unsigned) {
                            options.min = 0;
                            options.max = 2147483647;
                        }
                        else {
                            options.min = -2147483648;
                            options.max = 2147483647;
                        }
                        break;
                    case 'decimal':
                    case 'dec':
                    case 'float':
                    case 'double':
                        if (options.unsigned) {
                            options.min = 0;
                            options.max = 2147483647;
                        }
                        else {
                            options.min = -2147483648;
                            options.max = 2147483647;
                        }
                        break;
                    case 'date':
                    case 'datetime':
                    case 'timestamp':
                        options.minDate = '01-01-1970';
                        options.maxDate = undefined;
                        break;
                    case 'time':
                        break;
                    case 'year':
                        options.min = 1901;
                        options.max = 2155;
                        break;
                    case 'varchar':
                    case 'char':
                    case 'binary':
                    case 'varbinary':
                    case 'tinyblob':
                    case 'text':
                    case 'mediumtext':
                    case 'longtext':
                    case 'blob':
                    case 'mediumblob': // 16777215
                    case 'longblob': // 4,294,967,295
                    case 'set':
                    case 'enum':
                        break;
                }
                return {
                    name: column.COLUMN_NAME,
                    generator: column.DATA_TYPE,
                    options,
                };
            });
        });
    }
    orderTablesByForeignKeys(tables) {
        let sortedTables = [];
        const recursive = (branch) => {
            const table = branch[branch.length - 1];
            while (table.referencedTables.length > 0) {
                const tableName = table.referencedTables.pop();
                const referencedTable = tables.find((t) => {
                    return t.name === tableName;
                });
                if (referencedTable)
                    recursive([].concat(branch, referencedTable));
            }
            ;
            if (sortedTables.find((t) => t.name.toLowerCase() === table.name.toLowerCase()))
                return;
            sortedTables.push({
                name: table.name,
                disableTriggers: table.disableTriggers,
                maxLines: table.maxLines,
                addLines: table.addLines,
                columns: table.columns,
                before: table.before,
                after: table.after,
                referencedTables: []
            });
            branch.pop();
            return;
        };
        tables.forEach((table) => {
            recursive([table]);
        });
        return sortedTables;
    }
    ;
    generateJson(tables) {
        return {
            settings: {
                engine: this.customSchema.settings.engine,
                disableTriggers: !!this.customSchema.settings.disableTriggers,
                ignoredTables: this.customSchema.settings.ignoredTables,
                tablesToFill: this.customSchema.settings.tablesToFill,
                options: this.customSchema.settings.options,
                values: this.customSchema.settings.values,
            },
            tables: tables,
        };
    }
}
exports.Analyser = Analyser;
