"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const randomizer_1 = require("./randomizer");
const random_js_1 = require("random-js");
class TableService {
    constructor(dbConnector, maxCharLength, values) {
        this.dbConnector = dbConnector;
        this.maxCharLength = maxCharLength;
        this.values = values;
    }
    empty(table) {
        return __awaiter(this, void 0, void 0, function* () {
            console.log('empty: ', table.name);
            yield this.dbConnector.emptyTable(table);
        });
    }
    getForeignKeyValues(table, tableForeignKeyValues = {}, runRows) {
        return __awaiter(this, void 0, void 0, function* () {
            for (var c = 0; c < table.columns.length; c++) {
                const column = table.columns[c];
                if (column.foreignKey) {
                    const foreignKey = column.foreignKey;
                    let values = yield this.dbConnector.getValuesForForeignKeys(table.name, column.name, column.foreignKey.table, column.foreignKey.column, runRows, column.options.unique, column.foreignKey.where);
                    if (values.length === 0 && !column.options.nullable) {
                        throw new Error(`${table}: Not enough values available for foreign key ${foreignKey.table}.${foreignKey.column}`);
                    }
                    tableForeignKeyValues[`${column.name}_${foreignKey.table}_${foreignKey.column}`] = values;
                }
            }
        });
    }
    fill(table, reset) {
        return __awaiter(this, void 0, void 0, function* () {
            if (reset)
                yield this.empty(table);
            console.log('fill: ', table.name);
            yield this.before(table);
            yield this.generateData(table);
            yield this.after(table);
        });
    }
    before(table) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!table.before)
                return;
            for (const query of table.before) {
                yield this.dbConnector.executeRawQuery(query);
            }
        });
    }
    generateData(table) {
        return __awaiter(this, void 0, void 0, function* () {
            const tableForeignKeyValues = {};
            let previousRunRows = -1;
            let currentNbRows = yield this.dbConnector.countLines(table);
            batch: while (currentNbRows < table.lines) {
                previousRunRows = currentNbRows;
                const rows = [];
                const runRows = Math.min(1000, table.lines - currentNbRows);
                try {
                    yield this.getForeignKeyValues(table, tableForeignKeyValues, runRows);
                }
                catch (ex) {
                    console.warn(ex.message);
                    break batch;
                }
                for (let i = 0; i < runRows; i++) {
                    const row = {};
                    for (var c = 0; c < table.columns.length; c++) {
                        const column = table.columns[c];
                        if (column.options.autoIncrement)
                            continue;
                        if (column.values) {
                            if (Array.isArray(column.values)) {
                                row[column.name] = column.values[randomizer_1.Randomizer.randomInt(0, column.values.length - 1)];
                            }
                            else {
                                row[column.name] = this.values[column.values][randomizer_1.Randomizer.randomInt(0, this.values[column.values].length - 1)];
                            }
                            continue;
                        }
                        if (column.foreignKey) {
                            const foreignKeys = tableForeignKeyValues[`${column.name}_${column.foreignKey.table}_${column.foreignKey.column}`];
                            row[column.name] = foreignKeys[i];
                            continue;
                        }
                        switch (column.generator) {
                            case 'bit':
                                row[column.name] = randomizer_1.Randomizer.randomBit(column.options.max);
                                if (column.options.nullable && Math.random() <= 0.1)
                                    row[column.name] = null;
                                break;
                            case 'tinyint':
                                if (column.options.unsigned) {
                                    row[column.name] = randomizer_1.Randomizer.randomInt(0, 255);
                                }
                                else {
                                    row[column.name] = randomizer_1.Randomizer.randomInt(-128, 127);
                                }
                                if (column.options.nullable && Math.random() <= 0.1)
                                    row[column.name] = null;
                                break;
                            case 'bool':
                            case 'boolean':
                                row[column.name] = randomizer_1.Randomizer.randomInt(0, 1);
                                if (column.options.nullable && Math.random() <= 0.1)
                                    row[column.name] = null;
                                break;
                            case 'smallint':
                                if (column.options.unsigned) {
                                    row[column.name] = randomizer_1.Randomizer.randomInt(0, 65535);
                                }
                                else {
                                    row[column.name] = randomizer_1.Randomizer.randomInt(-32768, 32767);
                                }
                                if (column.options.nullable && Math.random() <= 0.1)
                                    row[column.name] = null;
                                break;
                            case 'mediumint':
                                if (column.options.unsigned) {
                                    row[column.name] = randomizer_1.Randomizer.randomInt(0, 16777215);
                                }
                                else {
                                    row[column.name] = randomizer_1.Randomizer.randomInt(-8388608, 8388607);
                                }
                                if (column.options.nullable && Math.random() <= 0.1)
                                    row[column.name] = null;
                                break;
                            case 'int':
                            case 'integer':
                            case 'bigint':
                                if (column.options.unsigned) {
                                    row[column.name] = randomizer_1.Randomizer.randomInt((column.options.min !== undefined ? column.options.min : 0), (column.options.max !== undefined ? column.options.max : 2147483647));
                                }
                                else {
                                    row[column.name] = randomizer_1.Randomizer.randomInt((column.options.min !== undefined ? column.options.min : -2147483648), (column.options.max !== undefined ? column.options.max : 2147483647));
                                }
                                if (column.options.nullable && Math.random() <= 0.1)
                                    row[column.name] = null;
                                break;
                            case 'decimal':
                            case 'dec':
                            case 'float':
                            case 'double':
                                if (column.options.unsigned) {
                                    row[column.name] = randomizer_1.Randomizer.randomFloat((column.options.min !== undefined ? column.options.min : 0), (column.options.max !== undefined ? column.options.max : 2147483647));
                                }
                                else {
                                    row[column.name] = randomizer_1.Randomizer.randomFloat((column.options.min !== undefined ? column.options.min : -2147483648), (column.options.max !== undefined ? column.options.max : 2147483647));
                                }
                                if (column.options.nullable && Math.random() <= 0.1)
                                    row[column.name] = null;
                                break;
                            case 'date':
                            case 'datetime':
                            case 'timestamp':
                                const min = column.options.min ? new Date(column.options.min) : new Date('01-01-1970');
                                const max = column.options.max ? new Date(column.options.max) : new Date();
                                row[column.name] = randomizer_1.Randomizer.randomDate(min, max);
                                if (column.options.nullable && Math.random() <= 0.1)
                                    row[column.name] = null;
                                break;
                            case 'time':
                                const hours = randomizer_1.Randomizer.randomInt(-838, +838);
                                const minutes = randomizer_1.Randomizer.randomInt(-0, +59);
                                const seconds = randomizer_1.Randomizer.randomInt(-0, +59);
                                row[column.name] = `${hours}:${minutes}:${seconds}`;
                                if (column.options.nullable && Math.random() <= 0.1)
                                    row[column.name] = null;
                                break;
                            case 'year':
                                row[column.name] = randomizer_1.Randomizer.randomInt(1901, 2155);
                                if (column.options.nullable && Math.random() <= 0.1)
                                    row[column.name] = null;
                                break;
                            case 'varchar':
                            case 'char':
                            case 'binary':
                            case 'varbinary':
                                if (column.options.max >= 36 && column.options.unique) {
                                    row[column.name] = random_js_1.uuid4(random_js_1.MersenneTwister19937.autoSeed());
                                }
                                else {
                                    row[column.name] = randomizer_1.Randomizer.randomString(randomizer_1.Randomizer.randomInt(column.options.min, Math.min(this.maxCharLength, column.options.max)));
                                    if (column.options.nullable && Math.random() <= 0.1)
                                        row[column.name] = null;
                                }
                                break;
                            case 'tinyblob':
                                row[column.name] = randomizer_1.Randomizer.randomString(randomizer_1.Randomizer.randomInt(0, 10));
                                if (column.options.nullable && Math.random() <= 0.1)
                                    row[column.name] = null;
                                break;
                            case 'text':
                            case 'mediumtext':
                            case 'longtext':
                                row[column.name] = randomizer_1.Randomizer.randomString(randomizer_1.Randomizer.randomInt(0, Math.min(this.maxCharLength, column.options.max)));
                                if (column.options.nullable && Math.random() <= 0.1)
                                    row[column.name] = null;
                                break;
                            case 'blob':
                            case 'mediumblob': // 16777215
                            case 'longblob': // 4,294,967,295
                                row[column.name] = randomizer_1.Randomizer.randomString(randomizer_1.Randomizer.randomInt(0, Math.min(this.maxCharLength, column.options.max)));
                                if (column.options.nullable && Math.random() <= 0.1)
                                    row[column.name] = null;
                                break;
                            case 'set':
                                row[column.name] = randomizer_1.Randomizer.randomBit(column.options.max);
                                if (column.options.nullable && Math.random() <= 0.1)
                                    row[column.name] = null;
                                break;
                            case 'enum':
                                if (column.options.nullable) {
                                    row[column.name] = Math.floor(Math.random() * (column.options.max + 1));
                                }
                                else {
                                    row[column.name] = Math.floor(Math.random() * (column.options.max)) + 1;
                                }
                                break;
                        }
                    }
                    rows.push(row);
                }
                currentNbRows += yield this.dbConnector.insert(table.name, rows);
                if (previousRunRows === currentNbRows) {
                    console.warn(`Last run didn't insert any new rows in ${table.name}`);
                    break batch;
                }
                console.log(currentNbRows + ' / ' + table.lines);
            }
        });
    }
    after(table) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!table.after)
                return;
            for (const query of table.after) {
                yield this.dbConnector.executeRawQuery(query);
            }
        });
    }
}
exports.TableService = TableService;
